import multiprocessing as mp
from concurrent.futures import ProcessPoolExecutor
from contextlib import asynccontextmanager
from functools import partial
from logging import getLogger
from typing import AsyncIterator, Callable, Tuple

import apluggy
from tblib import pickling_support

from nextline import spawned
from nextline.spawned import Command, QueueIn, QueueOut, RunArg, RunResult
from nextline.utils import MultiprocessingLogging, RunningProcess, run_in_process

from .monitor import Monitor

pickling_support.install()


def _call_all(*funcs) -> None:
    '''Execute callables and ignore return values.

    Used to call multiple initializers in ProcessPoolExecutor.
    '''
    for func in funcs:
        func()


@asynccontextmanager
async def run_session(
    hook: apluggy.PluginManager, run_arg: RunArg
) -> AsyncIterator[Tuple[RunningProcess[RunResult], Callable[[Command], None]]]:
    mp_context = mp.get_context('spawn')
    queue_in: QueueIn = mp_context.Queue()
    queue_out: QueueOut = mp_context.Queue()
    send_command = SendCommand(queue_in)
    monitor = Monitor(hook, queue_out)
    async with MultiprocessingLogging(mp_context=mp_context) as mp_logging:
        initializer = partial(
            _call_all,
            mp_logging.initializer,
            partial(spawned.set_queues, queue_in, queue_out),
        )
        executor_factory = partial(
            ProcessPoolExecutor,
            max_workers=1,
            mp_context=mp_context,
            initializer=initializer,
        )
        func = partial(spawned.main, run_arg)
        async with monitor:
            running = await run_in_process(func, executor_factory)
            yield running, send_command


def SendCommand(queue_in: QueueIn) -> Callable[[Command], None]:
    def _send_command(command: Command) -> None:
        logger = getLogger(__name__)
        logger.debug(f'send_pdb_command({command!r}')
        queue_in.put(command)

    return _send_command



import asyncio
import time
from logging import getLogger

from apluggy import PluginManager

from nextline import spawned

from ....spawned import QueueOut
from ....utils import to_thread

# from rich import print


class Monitor:
    def __init__(self, hook: PluginManager, queue: QueueOut):
        self._hook = hook
        self._queue = queue
        self._logger = getLogger(__name__)

    async def open(self):
        self._task = asyncio.create_task(self._monitor())

    async def close(self) -> None:
        up_to = 0.05
        start = time.process_time()
        while not self._queue.empty() and time.process_time() - start < up_to:
            await asyncio.sleep(0)
        await to_thread(self._queue.put, None)
        await self._task

    async def __aenter__(self):
        await self.open()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        del exc_type, exc_value, traceback
        await self.close()

    async def _monitor(self) -> None:
        while (event := await to_thread(self._queue.get)) is not None:
            await self._on_event(event)

    async def _on_event(self, event: spawned.Event) -> None:
        if isinstance(event, spawned.OnStartTrace):
            await self._hook.ahook.on_start_trace(event=event)
        elif isinstance(event, spawned.OnEndTrace):
            await self._hook.ahook.on_end_trace(event=event)
        elif isinstance(event, spawned.OnStartTraceCall):
            await self._hook.ahook.on_start_trace_call(event=event)
        elif isinstance(event, spawned.OnEndTraceCall):
            await self._hook.ahook.on_end_trace_call(event=event)
        elif isinstance(event, spawned.OnStartCmdloop):
            await self._hook.ahook.on_start_cmdloop(event=event)
        elif isinstance(event, spawned.OnEndCmdloop):
            await self._hook.ahook.on_end_cmdloop(event=event)
        elif isinstance(event, spawned.OnStartPrompt):
            await self._hook.ahook.on_start_prompt(event=event)
        elif isinstance(event, spawned.OnEndPrompt):
            await self._hook.ahook.on_end_prompt(event=event)
        elif isinstance(event, spawned.OnWriteStdout):
            await self._hook.ahook.on_write_stdout(event=event)
        else:
            self._logger.warning(f'Unknown event: {event!r}')



from logging import getLogger

from apluggy import PluginManager

from . import plugins, spec


def build_hook() -> PluginManager:
    hook = PluginManager(spec.PROJECT_NAME)
    hook.add_hookspecs(spec)

    plugins.register(hook)

    logger = getLogger(__name__)
    plugin_names = (f'{n!r}' for n, p in hook.list_name_plugin() if p)
    msg = f'Loaded plugins ({spec.PROJECT_NAME!r}): {",".join(plugin_names)}.'
    logger.info(msg)

    return hook



__all__ = ['register']


from apluggy import PluginManager

from .compose import CallableComposer
from .concurrency import TaskAndThreadKeeper, TaskOrThreadToTraceMapper
from .filter import FilerByModule, FilterByModuleName, FilterLambda
from .global_ import GlobalTraceFunc, TraceFuncCreator
from .local_ import LocalTraceFunc, TraceCallHandler
from .pdb_ import PdbInstanceFactory, Prompt
from .peek import PeekStdout
from .repeat import Repeater


def register(hook: PluginManager) -> None:
    hook.register(Repeater)
    hook.register(PeekStdout)
    hook.register(Prompt)
    hook.register(PdbInstanceFactory)
    hook.register(TraceCallHandler)
    hook.register(LocalTraceFunc)
    hook.register(TaskOrThreadToTraceMapper)
    hook.register(TaskAndThreadKeeper)
    hook.register(FilerByModule)
    hook.register(FilterLambda)
    hook.register(FilterByModuleName)
    hook.register(GlobalTraceFunc)
    hook.register(TraceFuncCreator)
    hook.register(CallableComposer)



import dataclasses
import datetime
from typing import Optional

from nextline import spawned
from nextline.plugin.spec import hookimpl
from nextline.spawned import RunArg
from nextline.types import RunInfo, RunNo
from nextline.utils import ExitedProcess
from nextline.utils.pubsub.broker import PubSub


class RunInfoRegistrar:
    def __init__(self) -> None:
        self._run_no: Optional[RunNo] = None
        self._script: Optional[str] = None
        self._run_info: Optional[RunInfo] = None

    @hookimpl
    def init(self, registry: PubSub) -> None:
        self._registry = registry

    @hookimpl
    async def on_change_script(self, script: str) -> None:
        self._script = script

    @hookimpl
    async def on_initialize_run(self, run_arg: RunArg) -> None:
        self._run_no = run_arg.run_no
        self._run_info = RunInfo(
            run_no=run_arg.run_no, state='initialized', script=self._script
        )
        await self._registry.publish('run_info', self._run_info)

    @hookimpl
    async def on_start_run(self) -> None:
        assert self._run_info is not None
        self._run_info = dataclasses.replace(
            self._run_info,
            state='running',
            started_at=datetime.datetime.utcnow(),
        )
        await self._registry.publish('run_info', self._run_info)

    @hookimpl
    async def on_end_run(
        self, exited_process: ExitedProcess[spawned.RunResult]
    ) -> None:
        assert self._run_info is not None
        run_result = exited_process.returned or spawned.RunResult(ret=None, exc=None)

        self._run_info = dataclasses.replace(
            self._run_info,
            state='finished',
            result=run_result.fmt_ret,
            exception=run_result.fmt_exc,
            ended_at=datetime.datetime.utcnow(),
        )
        await self._registry.publish('run_info', self._run_info)

        self._run_info = None
        self._run_no = None



import json
import traceback
from dataclasses import dataclass, field
from pathlib import Path
from queue import Queue
from types import CodeType, FrameType
from typing import Any, Callable, Optional, Tuple, Union

from typing_extensions import TypeAlias

from nextline.types import RunNo

from .commands import Command
from .events import Event

# if TYPE_CHECKING:
#     from sys import TraceFunction as TraceFunc  # type: ignore  # noqa: F401

# NOTE: TraceFunction from sys does not work well.
TraceFunction: TypeAlias = Callable[[FrameType, str, Any], "TraceFunction | None"]

TraceArgs: TypeAlias = Tuple[FrameType, str, Any]

QueueIn: TypeAlias = 'Queue[Command]'
QueueOut: TypeAlias = 'Queue[Event]'

Statement: TypeAlias = Union[str, Path, CodeType, Callable[[], Any]]


@dataclass
class RunArg:
    run_no: RunNo
    statement: Statement
    filename: Optional[str] = None


@dataclass
class RunResult:
    ret: Optional[Any]
    exc: Optional[BaseException]
    _fmt_ret: Optional[str] = field(init=False, repr=False, default=None)
    _fmt_exc: Optional[str] = field(init=False, repr=False, default=None)

    @property
    def fmt_ret(self) -> str:
        if self._fmt_ret is None:
            self._fmt_ret = json.dumps(self.ret)
        return self._fmt_ret

    @property
    def fmt_exc(self) -> str:
        if self._fmt_exc is None:
            if self.exc is None:
                self._fmt_exc = ''
            else:
                self._fmt_exc = ''.join(
                    traceback.format_exception(
                        type(self.exc),
                        self.exc,
                        self.exc.__traceback__,
                    )
                )
        return self._fmt_exc

    def result(self) -> Any:
        if self.exc is not None:
            # TODO: add a test for the exception
            raise self.exc
        return self.ret



import asyncio
import dataclasses
from logging import getLogger
from typing import Dict, Optional, Set

from nextline.plugin.spec import hookimpl
from nextline.spawned import (
    OnEndPrompt,
    OnEndTrace,
    OnEndTraceCall,
    OnStartPrompt,
    OnStartTrace,
    OnStartTraceCall,
    RunArg,
)
from nextline.types import PromptInfo, PromptNo, RunNo, TraceNo
from nextline.utils.pubsub.broker import PubSub


class PromptInfoRegistrar:
    def __init__(self) -> None:
        self._run_no: Optional[RunNo] = None
        self._last_prompt_frame_map: Dict[TraceNo, int] = {}
        self._trace_call_map: Dict[TraceNo, OnStartTraceCall] = {}
        self._prompt_info_map: Dict[PromptNo, PromptInfo] = {}
        self._keys: Set[str] = set()
        self._logger = getLogger(__name__)

    @hookimpl
    def init(self, registry: PubSub) -> None:
        self._registry = registry

    @hookimpl
    async def start(self) -> None:
        self._lock = asyncio.Lock()
        pass

    @hookimpl
    async def on_initialize_run(self, run_arg: RunArg) -> None:
        self._run_no = run_arg.run_no
        self._last_prompt_frame_map.clear()
        self._trace_call_map.clear()
        self._prompt_info_map.clear()
        self._keys.clear()

    @hookimpl
    async def on_end_run(self) -> None:
        async with self._lock:
            while self._keys:
                # the process might have been killed.
                key = self._keys.pop()
                await self._registry.end(key)

        self._run_no = None

    @hookimpl
    async def on_start_trace(self, event: OnStartTrace) -> None:
        assert self._run_no is not None
        trace_no = event.trace_no

        # TODO: Putting a prompt info for now because otherwise tests get stuck
        # sometimes for an unknown reason. Need to investigate
        prompt_info = PromptInfo(
            run_no=self._run_no,
            trace_no=trace_no,
            prompt_no=PromptNo(-1),
            open=False,
        )
        key = f"prompt_info_{trace_no}"
        async with self._lock:
            self._keys.add(key)
            await self._registry.publish(key, prompt_info)

    @hookimpl
    async def on_end_trace(self, event: OnEndTrace) -> None:
        trace_no = event.trace_no
        key = f"prompt_info_{trace_no}"
        async with self._lock:
            if key in self._keys:
                self._keys.remove(key)
                await self._registry.end(key)

    @hookimpl
    async def on_start_trace_call(self, event: OnStartTraceCall) -> None:
        self._trace_call_map[event.trace_no] = event

    @hookimpl
    async def on_end_trace_call(self, event: OnEndTraceCall) -> None:
        assert self._run_no is not None
        trace_no = event.trace_no
        trace_call = self._trace_call_map.pop(event.trace_no, None)
        if trace_call is None:
            self._logger.warning(f'No start event for {event}')
            return
        if not trace_call.frame_object_id == self._last_prompt_frame_map.get(trace_no):
            return

            # TODO: Sending a prompt info with "open=False" for now so that the
            #       arrow in the web UI moves when the Pdb is "continuing."

            # TODO: Add a test. Currently, the tests might pass without sending this
            #       prompt info.

        prompt_info = PromptInfo(
            run_no=self._run_no,
            trace_no=trace_no,
            prompt_no=PromptNo(-1),
            open=False,
            event=trace_call.call_event,
            file_name=trace_call.file_name,
            line_no=trace_call.line_no,
            trace_call_end=True,
        )
        await self._registry.publish('prompt_info', prompt_info)

        key = f"prompt_info_{trace_no}"
        async with self._lock:
            self._keys.add(key)
            await self._registry.publish(key, prompt_info)

    @hookimpl
    async def on_start_prompt(self, event: OnStartPrompt) -> None:
        assert self._run_no is not None
        trace_no = event.trace_no
        prompt_no = event.prompt_no
        trace_call = self._trace_call_map[trace_no]
        prompt_info = PromptInfo(
            run_no=self._run_no,
            trace_no=trace_no,
            prompt_no=prompt_no,
            open=True,
            event=trace_call.call_event,
            file_name=trace_call.file_name,
            line_no=trace_call.line_no,
            stdout=event.prompt_text,
            started_at=event.started_at,
        )
        self._prompt_info_map[prompt_no] = prompt_info
        self._last_prompt_frame_map[trace_no] = trace_call.frame_object_id

        await self._registry.publish('prompt_info', prompt_info)

        key = f"prompt_info_{trace_no}"
        async with self._lock:
            self._keys.add(key)
            await self._registry.publish(key, prompt_info)

    @hookimpl
    async def on_end_prompt(self, event: OnEndPrompt) -> None:
        trace_no = event.trace_no
        prompt_no = event.prompt_no
        prompt_info = self._prompt_info_map.pop(prompt_no)
        prompt_info_end = dataclasses.replace(
            prompt_info,
            open=False,
            command=event.command,
            ended_at=event.ended_at,
        )

        await self._registry.publish('prompt_info', prompt_info_end)

        key = f"prompt_info_{trace_no}"
        async with self._lock:
            self._keys.add(key)
            await self._registry.publish(key, prompt_info_end)



from __future__ import annotations

from concurrent.futures import ThreadPoolExecutor
from logging import getLogger
from queue import Queue
from typing import Callable, Iterator, MutableMapping, TypeVar

from apluggy import PluginManager, contextmanager
from typing_extensions import TypeAlias

from nextline.spawned.commands import PdbCommand
from nextline.spawned.plugin.spec import hookimpl
from nextline.spawned.types import QueueIn
from nextline.types import PromptNo, TraceNo

QueueMap: TypeAlias = MutableMapping[TraceNo, 'Queue[PdbCommand]']


class Prompt:
    '''A plugin that responds to the hook prompt() with commands from a queue.'''

    def __init__(self) -> None:
        self._logger = getLogger(__name__)

    @hookimpl
    def init(self, hook: PluginManager, queue_in: QueueIn) -> None:
        self._hook = hook
        self._queue_in = queue_in
        self._queue_map: QueueMap = {}

    @hookimpl
    @contextmanager
    def context(self) -> Iterator[None]:
        with relay_commands(self._queue_in, self._queue_map):
            yield

    @hookimpl
    def on_start_trace(self, trace_no: TraceNo) -> None:
        self._queue_map[trace_no] = Queue()

    @hookimpl
    def on_end_trace(self, trace_no: TraceNo) -> None:
        del self._queue_map[trace_no]

    @hookimpl
    def prompt(self, prompt_no: PromptNo) -> str:
        trace_no = self._hook.hook.current_trace_no()
        self._logger.debug(f'PromptNo: {prompt_no}')
        queue = self._queue_map[trace_no]

        while True:
            pdb_command = queue.get()
            try:
                assert pdb_command.trace_no == trace_no
            except AssertionError:
                msg = f'TraceNo mismatch: {pdb_command.trace_no} != {trace_no}'
                self._logger.exception(msg)
                raise
            if not (n := pdb_command.prompt_no) == prompt_no:
                self._logger.warning(f'PromptNo mismatch: {n} != {prompt_no}')
                continue
            return pdb_command.command


@contextmanager
def relay_commands(queue_in: QueueIn, queue_map: QueueMap):
    '''Pass the Pdb commands from the main process to the Pdb instances.'''
    logger = getLogger(__name__)

    def fn() -> None:
        assert queue_in
        while msg := queue_in.get():
            logger.debug(f'queue_in.get() -> {msg!r}')
            if isinstance(msg, PdbCommand):
                queue_map[msg.trace_no].put(msg)

    with ThreadPoolExecutor(max_workers=1) as executor:
        future = executor.submit(try_again_on_error, fn)  # type: ignore
        try:
            yield
        finally:
            queue_in.put(None)  # type: ignore
            future.result()


_T = TypeVar('_T')


def try_again_on_error(func: Callable[[], _T]) -> _T:
    '''Keep trying until the function succeeds without an exception.'''
    while True:
        try:
            return func()
        # except KeyboardInterrupt:
        #     raise
        except BaseException:
            logger = getLogger(__name__)
            logger.exception('')



from logging import getLogger

from apluggy import PluginManager

from nextline.spawned.types import QueueIn, QueueOut, RunArg

from . import plugins, skip, spec


def Hook(run_arg: RunArg, queue_in: QueueIn, queue_out: QueueOut) -> PluginManager:
    '''Return a plugin manager with the plugins registered.'''

    hook = PluginManager(spec.PROJECT_NAME)
    hook.add_hookspecs(spec)

    plugins.register(hook)

    logger = getLogger(__name__)
    plugin_names = (f'{n!r}' for n, p in hook.list_name_plugin() if p)
    msg = f'Loaded plugins ({spec.PROJECT_NAME!r}): {",".join(plugin_names)}.'
    logger.info(msg)

    hook.hook.init(
        hook=hook,
        run_arg=run_arg,
        modules_to_skip=skip.MODULES_TO_SKIP,
        queue_in=queue_in,
        queue_out=queue_out,
    )

    return hook



from nextline.plugin.spec import hookimpl
from nextline.utils.pubsub.broker import PubSub


class ScriptRegistrar:
    @hookimpl
    def init(self, registry: PubSub) -> None:
        self._registry = registry

    @hookimpl
    async def on_change_script(self, script: str, filename: str) -> None:
        await self._registry.publish('statement', script)
        await self._registry.publish('script_file_name', filename)



from logging import getLogger
from typing import Dict, Optional

from nextline.plugin.spec import hookimpl
from nextline.spawned import OnEndTraceCall, OnStartPrompt, OnStartTraceCall, RunArg
from nextline.types import PromptNotice, RunNo, TraceNo
from nextline.utils.pubsub.broker import PubSub


class PromptNoticeRegistrar:
    def __init__(self) -> None:
        self._run_no: Optional[RunNo] = None
        self._trace_call_map: Dict[TraceNo, OnStartTraceCall] = {}
        self._logger = getLogger(__name__)

    @hookimpl
    def init(self, registry: PubSub) -> None:
        self._registry = registry

    @hookimpl
    async def on_initialize_run(self, run_arg: RunArg) -> None:
        self._run_no = run_arg.run_no
        self._trace_call_map.clear()

    @hookimpl
    async def on_end_run(self) -> None:
        await self._registry.end('prompt_notice')
        self._run_no = None

    @hookimpl
    async def on_start_trace_call(self, event: OnStartTraceCall) -> None:
        self._trace_call_map[event.trace_no] = event

    @hookimpl
    async def on_end_trace_call(self, event: OnEndTraceCall) -> None:
        assert self._run_no is not None
        self._trace_call_map.pop(event.trace_no, None)

    @hookimpl
    async def on_start_prompt(self, event: OnStartPrompt) -> None:
        assert self._run_no is not None
        trace_no = event.trace_no
        prompt_no = event.prompt_no
        trace_call = self._trace_call_map[trace_no]
        prompt_notice = PromptNotice(
            started_at=event.started_at,
            run_no=self._run_no,
            trace_no=trace_no,
            prompt_no=prompt_no,
            prompt_text=event.prompt_text,
            event=trace_call.call_event,
            file_name=trace_call.file_name,
            line_no=trace_call.line_no,
        )
        await self._registry.publish('prompt_notice', prompt_notice)



from __future__ import annotations

from collections import defaultdict
from types import FrameType
from typing import Callable, DefaultDict, Dict, Optional, Set

from apluggy import PluginManager, contextmanager
from exceptiongroup import BaseExceptionGroup, catch

from nextline.spawned.plugin.spec import hookimpl
from nextline.spawned.types import RunResult, TraceArgs, TraceFunction
from nextline.spawned.utils import WithContext
from nextline.types import TraceNo


class LocalTraceFunc:
    '''A plugin that executes local trace functions.

    It calls different trace functions for each trace number. If a trace
    function doesn't exist for a trace number, this plugin creates a new one by
    calling the hook `create_local_trace_func`.

    A trace number is assigned to each async task or thread by another plugin.
    '''

    @hookimpl
    def init(self, hook: PluginManager) -> None:
        self._hook = hook
        factory = Factory(hook)
        self._map: DefaultDict[TraceNo, TraceFunction] = defaultdict(factory)

    @hookimpl
    def local_trace_func(self, frame: FrameType, event, arg) -> Optional[TraceFunction]:
        trace_no = self._hook.hook.current_trace_no()
        local_trace_func = self._map[trace_no]
        return local_trace_func(frame, event, arg)

    @hookimpl
    def finalize_run_result(self, run_result: RunResult) -> None:
        exc = run_result.exc
        if exc and exc.__traceback__ and isinstance(exc, KeyboardInterrupt):
            tb = exc.__traceback__
            while tb.tb_next:
                module = tb.tb_next.tb_frame.f_globals.get('__name__')
                if module == WithContext.__module__:
                    tb.tb_next = None
                    break
                tb = tb.tb_next


def Factory(hook: PluginManager) -> Callable[[], TraceFunction]:
    '''Return a function that creates a local trace function.'''

    def _factory() -> TraceFunction:
        trace = hook.hook.create_local_trace_func()

        @contextmanager
        def _context(frame, event, arg):
            '''A "with" block in which "trace" is called.'''

            keyboard_interrupt_raised = False

            def _keyboard_interrupt(exc: BaseExceptionGroup) -> None:
                nonlocal keyboard_interrupt_raised
                keyboard_interrupt_raised = True

            with hook.with_.on_trace_call(trace_args=(frame, event, arg)):
                with catch({KeyboardInterrupt: _keyboard_interrupt}):
                    # TODO: Using exceptiongroup.catch() for Python 3.8, 3.9, and 3.10.
                    #       Rewrite with except* for Python 3.11.
                    #       https://pypi.org/project/exceptiongroup/

                    yield

            if keyboard_interrupt_raised:
                # Reraise after the "with" block so that gen.throw() is not called.
                raise KeyboardInterrupt

        return WithContext(trace, context=_context)

    return _factory


class TraceCallHandler:
    '''A plugin that keeps the trace call arguments during trace calls.

    This plugin collect the trace call arguments when the context manager hook
    `on_trace_call` is entered. It responds to the first result only hooks
    `is_on_trace_call` and `current_trace_args`.
    '''

    def __init__(self) -> None:
        self._trace_args_map: Dict[TraceNo, TraceArgs] = {}
        self._traces_on_call: Set[TraceNo] = set()

    @hookimpl
    def init(self, hook: PluginManager) -> None:
        self._hook = hook

    @hookimpl
    @contextmanager
    def on_trace_call(self, trace_args: TraceArgs):
        trace_no = self._hook.hook.current_trace_no()
        self._traces_on_call.add(trace_no)
        self._trace_args_map[trace_no] = trace_args
        try:
            yield
        finally:
            self._traces_on_call.remove(trace_no)
            del self._trace_args_map[trace_no]

    @hookimpl
    def is_on_trace_call(self) -> Optional[bool]:
        trace_no = self._hook.hook.current_trace_no()
        return trace_no in self._traces_on_call

    @hookimpl
    def current_trace_args(self) -> Optional[TraceArgs]:
        trace_no = self._hook.hook.current_trace_no()
        return self._trace_args_map.get(trace_no)



import asyncio
from typing import Any, Optional

from transitions import EventData

from nextline.plugin import build_hook
from nextline.spawned import Command, Statement
from nextline.utils.pubsub.broker import PubSub

from .factory import build_state_machine


class Machine:
    '''The finite state machine of the nextline states.'''

    def __init__(self, run_no_start_from: int, statement: Statement):
        self.registry = PubSub[Any, Any]()
        self._hook = build_hook()
        self._hook.hook.init(
            hook=self._hook,
            registry=self.registry,
            run_no_start_from=run_no_start_from,
            statement=statement,
        )

        self._machine = build_state_machine(model=self)
        self._machine.after_state_change = self.after_state_change.__name__  # type: ignore

        assert self.state  # type: ignore

    def __repr__(self):
        # e.g., "<Machine 'running'>"
        return f'<{self.__class__.__name__} {self.state!r}>'

    async def after_state_change(self, event: EventData) -> None:
        if not (event.transition and event.transition.dest):
            # internal transition
            return
        await self._hook.ahook.on_change_state(state_name=self.state)  # type: ignore

    async def on_exit_created(self, _: EventData) -> None:
        await self._hook.ahook.start()

    async def on_enter_initialized(self, _: EventData) -> None:
        self._run_arg = self._hook.hook.compose_run_arg()
        await self._hook.ahook.on_initialize_run(run_arg=self._run_arg)

    async def on_enter_running(self, _: EventData) -> None:
        self.run_finished = asyncio.Event()
        run_started = asyncio.Event()

        async def run() -> None:
            async with self._hook.awith.run():
                run_started.set()
            await self.finish()  # type: ignore
            self.run_finished.set()

        self._task = asyncio.create_task(run())
        await run_started.wait()

    async def send_command(self, command: Command) -> None:
        await self._hook.ahook.send_command(command=command)

    async def interrupt(self) -> None:
        await self._hook.ahook.interrupt()

    async def terminate(self) -> None:
        await self._hook.ahook.terminate()

    async def kill(self) -> None:
        await self._hook.ahook.kill()

    async def on_close_while_running(self, _: EventData) -> None:
        await self.run_finished.wait()

    async def wait(self) -> None:
        await self.run_finished.wait()

    async def on_exit_finished(self, _: EventData) -> None:
        await self._task

    def exception(self) -> Optional[BaseException]:
        return self._hook.hook.exception()

    def result(self) -> Any:
        return self._hook.hook.result()

    async def on_enter_closed(self, _: EventData) -> None:
        await self.registry.close()
        await self._hook.ahook.close()

    async def on_reset(self, event: EventData) -> None:
        # TODO: Check the arguments
        await self._hook.ahook.reset(*event.args, **event.kwargs)

    async def __aenter__(self):
        await self.initialize()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        del exc_type, exc_value, traceback
        await self.close()



from typing import Any, Callable, Optional

import apluggy

from nextline import spawned
from nextline.utils import ExitedProcess, RunningProcess
from nextline.utils.pubsub.broker import PubSub

PROJECT_NAME = 'nextline_main'


hookspec = apluggy.HookspecMarker(PROJECT_NAME)
hookimpl = apluggy.HookimplMarker(PROJECT_NAME)


@hookspec
def init(
    hook: apluggy.PluginManager,
    registry: PubSub,
    run_no_start_from: int,
    statement: spawned.Statement,
) -> None:
    pass


@hookspec
async def start() -> None:
    pass


@hookspec
async def close(exc_type=None, exc_value=None, traceback=None) -> None:
    pass


@hookspec
async def reset(
    run_no_start_from: Optional[int],
    statement: Optional[spawned.Statement],
) -> None:
    pass


@hookspec
async def on_change_state(state_name: str) -> None:
    pass


@hookspec
async def on_change_script(script: str, filename: str) -> None:
    pass


@hookspec(firstresult=True)
def compose_run_arg() -> Optional[spawned.RunArg]:
    pass


@hookspec
async def on_initialize_run(run_arg: spawned.RunArg) -> None:
    pass


@hookspec
@apluggy.asynccontextmanager
async def run():
    yield


@hookspec
async def on_start_run(
    running_process: RunningProcess[spawned.RunResult],
    send_command: Callable[[spawned.Command], None],
) -> None:
    pass


@hookspec
async def interrupt() -> None:
    pass


@hookspec
async def terminate() -> None:
    pass


@hookspec
async def kill() -> None:
    pass


@hookspec
async def send_command(command: spawned.Command) -> None:
    pass


@hookspec
async def on_end_run(exited_process: ExitedProcess[spawned.RunResult]) -> None:
    pass


@hookspec(firstresult=True)
def exception() -> Optional[BaseException]:
    pass


@hookspec(firstresult=True)
def result() -> Any:
    pass


@hookspec
async def on_start_trace(event: spawned.OnStartTrace) -> None:
    pass


@hookspec
async def on_end_trace(event: spawned.OnEndTrace) -> None:
    pass


@hookspec
async def on_start_trace_call(event: spawned.OnStartTraceCall) -> None:
    pass


@hookspec
async def on_end_trace_call(event: spawned.OnEndTraceCall) -> None:
    pass


@hookspec
async def on_start_cmdloop(event: spawned.OnStartCmdloop) -> None:
    pass


@hookspec
async def on_end_cmdloop(event: spawned.OnEndCmdloop) -> None:
    pass


@hookspec
async def on_start_prompt(event: spawned.OnStartPrompt) -> None:
    pass


@hookspec
async def on_end_prompt(event: spawned.OnEndPrompt) -> None:
    pass


@hookspec
async def on_write_stdout(event: spawned.OnWriteStdout) -> None:
    pass
